// Container module.
module SoS_Explorer.Objective

    // Constrain given architecture via chromosome fixing.
    let FixChromosome(characteristics : float[,], capabilities : bool[,],
                      feasibleInterfaces : bool[,], directedInterfaces : bool,
                      architecture : bool[]) =

        // Number of elements
        let numSystems         = Array2D.length1 characteristics
        let numCharacteristics = Array2D.length2 characteristics
        let numCapabilities    = Array2D.length2 capabilities

        // System indices
        let sys_ReportA      = 0
        let sys_ReportB      = 1
        let sys_ReportC      = 2
        let sys_ReportD      = 3
        let sys_VisualA      = 4
        let sys_VisualB      = 5
        let sys_VisualC      = 6
        let sys_VisualD      = 7
        let sys_VisualE      = 8
        let sys_ClusterA     = 9
        let sys_ClusterB     = 10
        let sys_ClusterC     = 11
        let sys_PMA          = 12
        let sys_PMB          = 13
        let sys_PMC          = 14
        let sys_InhouseA     = 15
        let sys_InhouseB     = 16
        let sys_InhouseC     = 17
        let sys_InhouseD     = 18
        let sys_InhouseE     = 19
        let sys_InhouseF     = 20
        let sys_ApplicationA = 21
        let sys_ApplicationB = 22
        let sys_ApplicationC = 23

        // Characteristic indices
        let char_Language      = 0
        let char_Cost          = 1
        let char_Performance   = 2
        let char_Scalability   = 3
        let char_HasRG         = 4
        let char_HasVisual     = 5
        let char_HasClustering = 6
        let char_HasPM         = 7
        let char_HasApp        = 8

        // Capability indices
        let cap_OOXML            = 0
        let cap_PDF              = 1
        let cap_Graph            = 2
        let cap_Flow             = 3
        let cap_Multidimensional = 4
        let cap_Geospatial       = 5
        let cap_BIRCH            = 6
        let cap_DBSCAN           = 7
        let cap_SVM              = 8
        let cap_GLM              = 9
        let cap_DBN              = 10
        let cap_Application      = 11

        // Define hasSystem() function
        // Returns whether the given system is specified in the chromosome
        let hasSystem i =
            // Return whether system is specified
            architecture.[i]

        // Define hasInterface() function
        // Returns whether the given interface is specified in the chromosome
        let hasInterface i j =
            // If same system
            if i = j then
                false
            else
                // Index of interface in chromosome
                let jj = if j < i then j + 1 else j
                let k = 
                    if directedInterfaces then
                        (numSystems - 1) * (i + 1) + jj
                    else
                        let m = (min i j) + 1
                        m * numSystems - m * (m - 1) / 2 + (abs(i - j)) - 1
                // Return whether interface is specified
                architecture.[k]

        // Define setSystem() function
        // Sets the system in the chromosome to the specified value
        let setSystem i value =
            Array.set architecture i value |> ignore

        // Define setInterface() function
        // Sets the interface in the chromosome to the specified value
        let setInterface i j value =
            // If not same system
            if i <> j then
                // Index of interface in chromosome
                let jj = if j < i then j + 1 else j
                let k = 
                    if directedInterfaces then
                        (numSystems - 1) * (i + 1) + jj
                    else
                        let m = (min i j) + 1
                        m * numSystems - m * (m - 1) / 2 + (abs(i - j)) - 1
                Array.set architecture k value |> ignore

        // Define gamma (required paths matrix)
        let gamma = array2D [ [false; false; true;  true;  true];
                              [false; false; true;  true;  true];
                              [true;  true;  false; false; true];
                              [true;  true;  false; false; true];
                              [true;  true;  true;  true;  true] ]

        // Define has path function
        let rec hasPath i j =
            hasInterface i j

        // Identify and add missing capabilities
        for i = 0 to numCapabilities - 1 do
            let mutable j = 0
            let mutable k = -1
            let mutable hasCapability = false
            while (not hasCapability) && (j < numSystems) do
                if capabilities.[j, i] then
                    if hasSystem j then
                        hasCapability <- true
                    else
                        k <- j
                j <- j + 1
            if (not hasCapability) && (k > -1) then
                setSystem k true

        // Remove infeasible interfaces from architecture
        for i = 0 to numSystems - 1 do
            for j = 0 to numSystems - 1 do
                // If different systems
                if i <> j then
                    // If interface is specified
                    if hasInterface i j then
                        // If infeasible, remove
                        if not (hasSystem i && hasSystem j && feasibleInterfaces.[i, j]) then
                            setInterface i j false

        // Add required interfaces to architecture
        for i = 0 to numSystems - 1 do
            if hasSystem i then
                for m = char_HasRG to char_HasApp do
                    if abs(characteristics.[i, m]) > 0.01 then
                        for n = char_HasRG to char_HasApp do
                            if ((m <> n) && gamma.[m - char_HasRG, n - char_HasRG]) then
                                for j = 0 to numSystems - 1 do
                                    if ((i <> j) && (hasSystem j) && (abs(characteristics.[j, n]) > 0.01)) then
                                        if not (hasPath i j) then
                                            setInterface i j true
